/***********************************************************************NTNU IT*
* AIX Speed Test using the power cpu performance counters. Requires bos.pmapi
* fileset to be installed on the target system.
*------------------------------------------------------------------------------
* $Id: TestSpeed-aix.c 69 2009-09-14 06:42:16Z joern $
* (c) 2009 jorn.amundsen@ntnu.no
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pmapi.h>
#include <limits.h>
#include <unistd.h>
#include <sys/processor.h>
#include <sys/utsname.h>

#if !defined(LLONG_MAX)
#define LLONG_MAX (~0ull>>1)
#endif

#include "SHA3api_ref.h"
#include "byteorder.h"

static int init_counters(int *i_cyc, int *i_inst, int *i_iop0, int *i_iop1);
static void do_header(char *line, FILE *fp, int argc, char **argv);
#define teeprint(s,fp) (fputs(s,stdout), fputs(s,fp))

#define ST_OUTFILE "SpeedTest.txt"
#define ST_FILE_LBUF 254
#define ST_MD_SIZE  64
#define ST_NUM_TESTRUNS 10000

/* cycle counter mask bits (FXUn is FXU0 or FXU1) */
#define HAS_PM_FXU_FIN       0x1
#define HAS_PM_CYC           0x2
#define HAS_PM_INST_CMPL     0x4
#define HAS_PM_RUN_CYC       0x8
#define HAS_PM_FXUn_FIN      0x10
#define HAS_PM_RUN_INST_CMPL 0x20
#define HAS_GROUP_VERIFIED   0x40


static int bitlens[] = {  /* the supported hash output sizes */
	224, 256, 384, 512, 0 
};
static int mdruns[] = {   /* standard and SUPERCOP benchmark message lengths */
	1, 8, 10, 64, 100, 576, 1000, 1536, 4096, 10000, 100000, 0
};
pm_data_t pmdata;
long long accu_min[MAX_COUNTERS];


int
main(int argc, char **argv)
{
	BitSequence *Msg, *MD;
	FILE *fp_out;
	char *line;
	unsigned mdsum;
	int i, rc, max_md, pri_mask, i_cyc, i_inst, i_iop0, i_iop1;

	/* Constrain the process to stay at one core */
	if (bindprocessor(BINDPROCESS, getpid(), mycpu()) == -1) {
		perror("\tbindprocessor()");
		exit(1);
	}

	/* find max message length and allocate speed test memory */
	max_md = mdruns[0];
	for (i = 0, max_md = mdruns[0]; mdruns[i] > 0; i++)
		if (mdruns[i] > max_md)
			max_md = mdruns[i];
	Msg = malloc(max_md+1);
	MD = malloc(ST_MD_SIZE);
	line = malloc(ST_FILE_LBUF+2);
	if (!Msg || !MD || !line) {
		fputs("No memory\n\n", stderr);
		exit(1);
	}

	/* open speed test output file */
	unlink(ST_OUTFILE);
	if ((fp_out = fopen(ST_OUTFILE, "w")) == NULL) {
		fputs("Couldn't open " ST_OUTFILE " for write\n", stderr);
		exit(1);
	}

	/* output header to terminal and file */
	do_header(line, fp_out, argc, argv);

	/* initialize message */
	for (i = 0; i < max_md; i++)
		Msg[i] = (BitSequence)i;

	/* initialize cycle counters */
	pri_mask = init_counters(&i_cyc, &i_inst, &i_iop0, &i_iop1);

	
	/* output size loop */
	for (mdsum = 0, i = 0; bitlens[i] > 0; i++) {
		int j;
		
		for(j = 0; mdruns[j] > 0; j++) {
			double ninv;
			int k, n;

			/* run the test NUM_TESTRUNS times and find minimum cycle count */
			n = mdruns[j];
			ninv = 1. / (double)n;
			accu_min[i_cyc] = LLONG_MAX;
			for(k = 0; k < ST_NUM_TESTRUNS; k++ ) {
				pm_start_mythread();
				Hash(bitlens[i], Msg, n*8, MD);
				pm_stop_mythread();
				if ((rc = pm_get_data_mythread(&pmdata)) != 0) {
					pm_error("pm_get_data_mythread", rc);
					exit(1);
				}
				if (pmdata.accu[i_cyc] < accu_min[i_cyc])
					memcpy(accu_min, pmdata.accu, sizeof(accu_min));
				pm_reset_data_mythread();
			}
			mdsum ^= *(unsigned *)MD;
			sprintf(line, "%6d  %8d  %8.3lf  %8.2lf  %8.2lf\n", bitlens[i], n,
				accu_min[i_cyc]*ninv, accu_min[i_inst]*ninv,
				(accu_min[i_iop0] + (pri_mask & HAS_PM_FXUn_FIN ?
				accu_min[i_iop1] : 0))*ninv);
			teeprint(line, fp_out);
		}
	}

	ld_swap32(&mdsum, mdsum);
	sprintf(line, "\n  mdsum 0x%x\n\n", mdsum);
	teeprint(line, fp_out);

	fclose(fp_out);
	return 0;
}


/* performance counter  table */
#define I_CYC           0
#define I_INST_CMPL     1
#define I_RUN_CYC       2
#define I_RUN_INST_CMPL 3
#define I_FXU0_FIN      4
#define I_FXU1_FIN      5
#define I_FXU_FIN       6
#define I_LAST I_FXU_FIN
static int pmc[I_LAST+1];


/*
 * Find the best suited counter group on the underlying processor to
 * count run cycles, run instructions and number of fixed point instructions.
 */

static int
init_counters(int *i_cyc, int *i_inst, int *i_iop0, int *i_iop1)
{
	pm_info2_t pminfo;
	pm_groups_info_t pmgroups;
	pm_prog_t pmprog;
	int filter, hwc, hwg, g_pri, rc;

	/* initialize the event counter and fill in processor events */
	filter = PM_UNVERIFIED | PM_CAVEAT | PM_VERIFIED | PM_GET_GROUPS;
	if ((rc = pm_initialize(filter, &pminfo, &pmgroups, PM_CURRENT)) != 0) {
		pm_error("pm_initialize", rc);
		exit(1);
	}
	if (pmgroups.event_groups == NULL) {
		fputs("pm_initialize: no groups\n", stderr);
		exit(1);
	}

	/* initialize the pmprog struct */
	for (hwc = 0; hwc < pminfo.maxpmcs; hwc++)
		pmprog.events[hwc] = COUNT_NOTHING;
	pmprog.mode.w = 0;
	pmprog.mode.b.user = 1;
	pmprog.mode.b.is_group = 1;

	/* loop through groups to find the group with the most useful counters */
	for (g_pri = hwg = 0; hwg < pmgroups.maxgroups; hwg++) {
		int pri, unverified;

		unverified = 1;
		for (pri = hwc = 0; hwc < pminfo.maxpmcs; hwc++) {
			int e_id;
			pm_events2_t *e;

			e_id = pmgroups.event_groups[hwg].events[hwc];
			if (e_id == COUNT_NOTHING || pminfo.maxevents[hwc] == 0)
				break;
			e = pminfo.list_events[hwc] + e_id;
			unverified = unverified && e->status.b.unverified;

			if (strcmp(e->short_name, "PM_RUN_CYC") == 0)
				pri |= HAS_PM_RUN_CYC, pmc[I_RUN_CYC] = hwc;
			else if (strcmp(e->short_name, "PM_CYC") == 0)
			   pri |= HAS_PM_CYC, pmc[I_CYC] = hwc;
			else if (strcmp(e->short_name, "PM_RUN_INST_CMPL") == 0)
			   pri |= HAS_PM_RUN_INST_CMPL, pmc[I_RUN_INST_CMPL] = hwc;
			else if (strcmp(e->short_name, "PM_INST_CMPL") == 0)
			   pri |= HAS_PM_INST_CMPL, pmc[I_INST_CMPL] = hwc;
			else if (strcmp(e->short_name, "PM_FXU0_FIN") == 0)
			   pri |= HAS_PM_FXUn_FIN, pmc[I_FXU0_FIN] = hwc;
			else if (strcmp(e->short_name, "PM_FXU1_FIN") == 0)
			   pri |= HAS_PM_FXUn_FIN, pmc[I_FXU1_FIN] = hwc;
			else if (strcmp(e->short_name, "PM_FXU_FIN") == 0)
			   pri |= HAS_PM_FXU_FIN, pmc[I_FXU_FIN] = hwc;
		}

		/* prefer PM_RUN_xxx over PM_xxx */
		if (pri & HAS_PM_RUN_CYC)
			pri &= ~HAS_PM_CYC;
		if (pri & HAS_PM_RUN_INST_CMPL)
			pri &= ~HAS_PM_INST_CMPL;

		/* prefer PM_FXUn_FIN over PM_FXU_FIN */
		if (pri & HAS_PM_FXUn_FIN)
			pri &= ~HAS_PM_FXU_FIN;

		/* prefer Verified and Caveat over Unverified */
		if (!unverified)
			pri |= HAS_GROUP_VERIFIED;

		if (pri > g_pri) {
			g_pri = pri;
			pmprog.events[0] = hwg;
		}
	}

	/* output final counter indices */
	*i_cyc = pmc[g_pri & HAS_PM_RUN_CYC ? I_RUN_CYC : I_CYC];
	*i_inst = pmc[g_pri & HAS_PM_RUN_INST_CMPL ? I_RUN_INST_CMPL : I_INST_CMPL];
	if (g_pri & HAS_PM_FXUn_FIN)
		*i_iop0 = pmc[I_FXU0_FIN], *i_iop1 = pmc[I_FXU1_FIN];
	else
		*i_iop0 = pmc[I_FXU0_FIN];

	/* initialize PM counter group */
	if ((rc = pm_set_program_mythread(&pmprog)) != 0) {
		pm_error("pm_set_program_mythread", rc);
		exit(1);
	}

	return g_pri;
}


#define ARGCMP(s, arg) (strncmp(s, arg, strlen(s)))

static void
do_header(char *line, FILE *fp, int argc, char **argv)
{
	/* print header text */
	teeprint("\nBlueMidnightWish performance in cycles/byte with various "
		"message byte lengths\n", fp);
	sprintf(line, "               [ minimum of %d, results in units "
		"per byte ]\n\n", ST_NUM_TESTRUNS);
	teeprint(line, fp);

	/* print test date */
	{
		time_t t;
		struct tm *tp;

		time(&t);
		tp = localtime(&t);
		if (tp)
			sprintf(line, "date:     %4d-%02d-%02d %02d:%02d:%02d\n",
				tp->tm_year+1900, tp->tm_mon+1, tp->tm_mday, tp->tm_hour,
				tp->tm_min,	tp->tm_sec);
		else
			strcpy(line, "date:     <unknown>\n");

		teeprint(line, fp);
	}

	/* print cmd line arguments */
	{
		char *abi, *arch;

		abi = arch = "<unknown>";
		for (++argv; *argv; argv++) {
			if (ARGCMP("abi=", *argv) == 0)
				abi = *argv + strlen("abi=");
			else if (ARGCMP("arch=", *argv) == 0)
				arch = *argv + strlen("arch=");
		}
		sprintf(line, "arch:     %s\nabi:      %s\n", arch, abi);
		teeprint(line, fp);
	}

	/* print name and kernel */
	{
		struct utsname uts;

		if (uname(&uts) == 0)
#if defined(_AIX)
			sprintf(line, "hostname: %s\nmachine:  %s\nosrel:    %s.%s\n",
				uts.nodename, uts.machine, uts.version, uts.release);
#else
			sprintf(line, "hostname: %s\nmachine:  %s\nosrel:    %s\n",
				uts.nodename, uts.machine, uts.release);
#endif /*_AIX*/
		else
			strcpy(line, "hostname: <unknown>\nmachine:  <unknown>\n"
				"osrel:    <unknown>\n");
		teeprint(line, fp);
	}

	/* find and print compiler version */
	{
#if defined(__SUNPRO_C)
		int major, minor, micro;
#endif /*__SUNPRO_C*/

#if defined(__VERSION__) && !defined(__ICC)
		sprintf(line, "compiler: gcc %s\n", __VERSION__);
#elif defined(__xlc__)
		sprintf(line, "compiler: xlc %s\n", __xlc__);
#elif defined(__ICC)
		sprintf(line, "compiler: icc %d.%d.%d\n", __ICC/100, __ICC%100,
			__INTEL_COMPILER_BUILD_DATE);
#elif defined(__PGIC__)
		sprintf(line, "pgcc %d.%d.%d\n", __PGIC__, __PGIC_MINOR__,
			__PGIC_PATCHLEVEL__);
#elif defined(__SUNPRO_C)
		micro = __SUNPRO_C & 0xf;
		minor = (__SUNPRO_C >> 4) & 0xf;
		major = (__SUNPRO_C >> 8) & 0xf;

		if (micro)
			sprintf(line, "compiler: sun_c %d.%d.%d", major, minor, micro);
		else
			sprintf(line, "compiler: sun_c %d.%d", major, minor);
#else
		strcpy(line, "compiler: <unknown>");
#endif

		teeprint(line, fp);
	}

	/* finally print bytes/cycles header */
#if defined(_AIX)
	teeprint("\nMDSIZE  NUMBYTES    CYCLES     INSTR    INTOPS\n", fp);
#else
	teeprint("\nMDSIZE  NUMBYTES    CYCLES\n", fp);
#endif /*_AIX*/
}
